var lWarn = {logln:function (o){}};
var lInfo = {logln:function (o){}};

var lInfoFILE_NAME = "sr-config.log";
lWarn = new Logger(lInfoFILE_NAME);
lInfo = new Logger(lInfoFILE_NAME);

function storeSubParameter(path, value) {
    Property.setProperty('entries/' + path, value);
    Property.setFlag('entries/' + path, 'ARCHIVE', true);
}

function getNewIdValue() {
    var idValue = 0;
    var baseNode = Property.getNode('entries');
    if (baseNode == null) {
        return idValue;
    }
    var oArrayNode = baseNode.getChildren();
    for ( var iIndex = 0; iIndex < oArrayNode.length; iIndex++) {
        var oChildNode = oArrayNode[iIndex];
        if (oChildNode.getChild('id') != null) {
            if (oChildNode.getChild('id').getValue() != null) {
                if (parseInt("" + oChildNode.getChild('id').getValue(), 10) >= idValue) {
                    idValue = parseInt("" + oChildNode.getChild('id').getValue(), 10) + 1;
                }
            }
        }
    }
    return idValue;
}

function isPropertyTreeNotCurrent(currentVersionString, compareVersionString) {
    var curVersion = currentVersionString.split('.'),
        comVersion = compareVersionString.split('.');
    for (var i=0; i < curVersion.length; i++) {
        if ((i < comVersion.length) &&
            (parseInt(curVersion[i], 10) < parseInt(comVersion[i], 10))) {
            return true;
        }
    }
    return false;
}

function convertOldTriggers() {
    var ndTriggers = Property.getNode('triggers');
    if (!ndTriggers) {
        return;
    }
    var triggers = ndTriggers.getChildren();
    var i;
    for (i = 0; i < triggers.length; ++i) {
        var ndOldParams = triggers[i].getChild('params'),
            oldParams,
            oldTrigger,
            oldActions;

        if (!ndOldParams) {
            lInfo.logln('Error: no old params found to convert on trigger ' + triggers[i].name);
            deleteOldTrigger(triggers[i]);
            continue;
        }

        try {
            // lInfo.logln('Debug: old params: ' + ndOldParams.getValue());
            oldParams = JSON.parse(ndOldParams.getValue());
            // lInfo.logln('Debug: old trigger: ' + decodeURIComponent(oldParams.trigger));
            oldTrigger = JSON.parse(oldParams.trigger);
            // lInfo.logln('Debug: old actions: ' + decodeURIComponent(oldParams.actions));
            oldActions = JSON.parse(oldParams.actions);
        } catch (e) {
            lInfo.logln('Deleting old trigger '+ triggers[i].name +': failed to parse');
            deleteOldTrigger(triggers[i]);
            continue;
        }

        //
        // validity checks
        //

        if (!oldTrigger
            || isNaN(oldTrigger.scene)
            || isNaN(oldTrigger.zone)
                // allowed to be unset in very old versions (< 0.1.4)
            || (oldTrigger.group && isNaN(oldTrigger.group))
        ) {
            // delete invalid triggers
            lInfo.logln('Deleting old trigger '+ triggers[i].name +': missing trigger');
            deleteOldTrigger(triggers[i]);
            continue;
        }

        if (!validScene(oldTrigger.scene)) {
            lInfo.logln('Deleting old trigger '+ triggers[i].name +': scene ' + oldTrigger.scene +' is not supported by this version');
            deleteOldTrigger(triggers[i]);
            continue;
        }

        if (!oldActions
            || !oldActions.actions
            || !(oldActions.actions.length > 0)
        ) {
            // delete invalid triggers
            lInfo.logln('Deleting old trigger '+ triggers[i].name +': missing actions');
            deleteOldTrigger(triggers[i]);
            continue;
        }
        oldActions = oldActions.actions;

        //
        // Conversions
        //

        // new params object
        var params = {
            name: triggers[i].name, // or oldParams.name
            disabled: false, // there existed no possibility to disable an action
            scene: parseInt(oldTrigger.scene, 10),
            zone: parseInt(oldTrigger.zone, 10)
        };

        // Convert delay
        if (oldParams.duration === undefined) {
            params.delay = 0;
        } else if (isNaN(oldParams.duration)) {
            lInfo.logln('Warning: invalid delay on trigger '+ triggers[i].name +', assuming 0');
            params.delay = 0;
        } else {
            params.delay = 60 * parseInt(oldParams.duration, 10);
        }

        // Convert trigger group
        if (isNaN(oldTrigger.group)) {
            lInfo.logln('Warning: no trigger group is set on '+ triggers[i].name +': defaulting to group 1 (Lights)');
            params.group = 1;
        } else {
            params.group = parseInt(oldTrigger.group, 10);
        }

        // Convert old actions
        params.actions = [];
        var a;
        for (a = 0; a < oldActions.length; ++a) {
            var p = {};
            if (isNaN(oldActions[a].zone)
                || (isNaN(oldActions[a].scene) && isNaN(oldActions[a].value))
                || (oldActions[a].group && isNaN(oldActions[a].group))
                || (!isNaN(oldActions[a].value)
                    && (oldActions[a].value < 0 || oldActions[a].value > 255))
            ) {
                lInfo.logln('Error: no or invalid action set on trigger '+ triggers[i].name +' action at index '+ a +'. Skipping');
                continue;
            }
            p.zone = parseInt(oldActions[a].zone, 10);
            if (oldActions[a].group === undefined) {
                lInfo.logln('Warning: no group is set on '+ triggers[i].name +' action at index '+ a +', defaulting to group 1 (Lights)');
                p.group = 1;
            } else {
                p.group = parseInt(oldActions[a].group, 10);
            }
            if (oldActions[a].scene !== undefined) {
                if (!validScene(oldActions[a].scene)) {
                    lInfo.logln('Warning unsupported scene '
                            + oldActions[a].scene +' call from action '+ a
                            +' on '+ triggers[i].name +': skipping');
                    continue;
                }

                p.scene = parseInt(oldActions[a].scene, 10);
            } else if (oldActions[a].value !== undefined) {
                p.outputValue = parseInt(oldActions[a].value, 10);
            }
            params.actions.push(p);
        }
        if (params.actions.length === 0) {
            lInfo.logln('Error: no valid actions on trigger '+ triggers[i].name +': deleting');
            deleteOldTrigger(triggers[i]);
            continue;
        }

        // store converted trigger
        deleteOldTrigger(triggers[i]);
        addTrigger(params);
        lInfo.logln('Converted trigger '+ params.idname);
    }
    Property.store();
} // convertOldTriggers

/**
 * Check if a scene is supported by this scene responder version
 * @param s {Number} Scene id to check
 * @return {Boolean} True if supported, false if unsupported
 */
function validScene(s) {
    if ((s >= 10 && s <= 16) || (s >= 32 && s <= 63) || s >= 74) {
        return false;
    }
    return true;
}

/**
 * Returns an unused unique identifier (id)
 * @return {String} A unique string
 */
function getId() {
    var ndTriggers = Property.getNode('triggers'),
        triggers,
        max = 0;

    if (ndTriggers) {
        triggers = ndTriggers.getChildren();
        while (ndTriggers.getChild(max.toString())) {
            ++max;
        }
        return max.toString();

    } else {
        return "0";
    }
}

/**
 * Checks if parameters conform to what is expected
 * @param triggerParams {Object} Object to check
 * @return {Object} Conform object or null on error
 */
function checkParams(triggerParams) {
    if (!triggerParams.name) {
        lInfo.logln('Error: trigger has no name');
        return null;
    }

    if (!triggerParams.actions || triggerParams.actions.length === 0) {
        lInfo.logln('Error: trigger has no actions');
        return null;
    }

    var actions = [];
    var i;
    for (i = 0; i < triggerParams.actions.length; ++i) {
        var a = triggerParams.actions[i];
        if (!a.device && isNaN(a.group)) { // isNan(undefined) === true
            lInfo.logln('Error: no or invalid group or device defined as target on action ' + i);
            return null;
        }
        if (a.blink === undefined && a.scene === undefined && a.outputValue === undefined) {
            lInfo.logln('Error: no action defined at index ' + i);
            return null;
        }
        if (
            (a.scene !== undefined && isNaN(a.scene))
            || (a.group !== undefined && isNaN(a.group))
            || (a.outputValue !== undefined && (
                isNaN(a.outputValue) || a.outputValue < 0 || a.outputValue > 255
            ))
        ) {
            lInfo.logln('Error: invalid action defined at index ' + i);
            return null;
        }

        var action = {};
        if (a.zone !== undefined) {
            action.zone = parseInt(a.zone, 10);
        }
        if (a.group !== undefined) {
            action.group = parseInt(a.group, 10);
        }
        if (a.device !== undefined) {
            action.device = a.device;
        }
        if (a.blink) {
            action.blink = true;
        }
        if (a.scene !== undefined) {
            action.scene = parseInt(a.scene, 10);
        }
        if (a.outputValue !== undefined) {
            action.outputValue = parseInt(a.outputValue, 10);
        }
        actions.push(action);
    }

    if (isNaN(triggerParams.delay)) {
        lInfo.logln('Warning: invalid delay on action ' + i +' assuming 0');
        triggerParams.delay = 0;
    }

    if (isNaN(triggerParams.scene)
        || isNaN(triggerParams.group)
        || isNaN(triggerParams.zone)
    ) {
        lInfo.logln('Error: invalid trigger on action ' + i);
        return null;
    }

    // create the valid params object
    var params = {
        name: triggerParams.name,
        delay: parseInt(triggerParams.delay, 10),
        scene: parseInt(triggerParams.scene, 10),
        group: parseInt(triggerParams.group, 10),
        zone: parseInt(triggerParams.zone, 10),
        disabled: !!triggerParams.disabled,
        actions: actions
    };
    // add id, idname fields if they were supplied
    if (triggerParams.id) {
        params.id = triggerParams.id;
        params.idname = params.id + ' (' + params.name + ')';
    }

    return params;
} // checkParams

/**
 * Updates a node if it exists or creates a new archived property node otherwise
 * @param path {String} Relative path
 * @param value {String|Boolean|Integer} Value to set
 */
function updateOrCreateNode(path, value) {
    var nd = Property.getNode(path);
    if (nd) {
        nd.setValue(value);
    } else {
        createNode(path, value);
    }
}

/**
 * Creates a new archived property node with given value
 * @param path {String} Relative path
 * @param value {String|Boolean|Integer} Value to set
 * @return {Object} The newly created node
 */
function createNode(path, value) {
    var nd = new Property(path);
    if (value !== undefined) {
        nd.setValue(value);
    }
    nd.setFlag('ARCHIVE', true);
    return nd;
}

/**
 * Stores _VALID_ parameters into the property tree.
 * Use checkParams(params) to validate them first!
 * Make sure to call Property.store() after calling this method
 * @param params {Object} Object to store
 * @return {Object} The trigger node
 */
function storeParams(params) {
    var triggerPath = 'triggers/' + params.id;
    var actionPath = triggerPath + '/actions';
    var i, c;

    var ndTrigger = Property.getNode(triggerPath) || createNode(triggerPath);

    updateOrCreateNode(triggerPath + '/id', params.id);
    updateOrCreateNode(triggerPath + '/name', params.name);
    updateOrCreateNode(triggerPath + '/delay', params.delay);
    updateOrCreateNode(triggerPath + '/disabled', !!params.disabled);
    updateOrCreateNode(triggerPath + '/scene', params.scene);
    updateOrCreateNode(triggerPath + '/group', params.group);
    updateOrCreateNode(triggerPath + '/zone', params.zone);

    // actions
    var ndActions = ndTrigger.getChild('actions');
    if (ndActions) {
        // clean out actions, as indices might not match up
        ndTrigger.removeChild(ndActions);
    }
    ndActions = createNode(actionPath);

    // populate actions
    for (i = 0; i < params.actions.length; ++i) {
        var p = actionPath + '/' + i;
        createNode(p);

        if (params.actions[i].zone !== undefined) {
            createNode(p + '/zone', params.actions[i].zone);
        }
        if (params.actions[i].group !== undefined) {
            createNode(p + '/group', params.actions[i].group);
        }
        if (params.actions[i].device) {
            createNode(p + '/device', params.actions[i].device);
        }
        if (params.actions[i].blink) {
            createNode(p + '/blink', true);
        }
        if (params.actions[i].scene !== undefined) {
            createNode(p + '/scene', params.actions[i].scene);
        }
        if (params.actions[i].outputValue !== undefined) {
            createNode(p + '/outputValue', params.actions[i].outputValue);
        }
    }
    return ndTrigger;
} // storeParams


/**
 * Add a new trigger
 * @param params {Object} The new trigger is created from these parameters
 */
function addTrigger(triggerParams) {
    // check triggerParams and set params for serialization
    var params = checkParams(triggerParams),
        // tmpId is used to communicate the id of the newly created trigger
        // to the client. Since there is no way to specify a return value
        // for an event, the client will be able to specify a temporary id
        // tmpId and poll for that id. When it's created, it'll contain the
        // definitive id of the created trigger.
        // The client is charged with deleting this node
        tmpId;

    if (params) {
        if (params.id === undefined || params.id.substr(0, 3) === 'tmp') {
            if (params.id !== undefined) {
                tmpId = params.id;
            }
            params.id = getId();
            params.idname = params.id + ' (' + params.name + ')';
        }

        storeParams(params);

        if (tmpId !== undefined) {
            var evt = new Event('scene-responder.create.' + tmpId, { id: params.id });
            evt.raise();
        }

        Property.store();
    }
}

/**
 * Deletes an old (and converted but not saved) trigger from the property tree
 */
function deleteOldTrigger(ndTrigger) {
    ndTrigger.getParent().removeChild(ndTrigger);
    Property.store();
}

function updateData() {
    var baseNodes = Property.getNode('triggers');
    lInfo.logln("UpdateDataStart");
    if (baseNodes == null) {
        lInfo.logln("nothing to Update");
        return;
    }
    var fError=false;
    var oBaseArray=baseNodes.getChildren();
    for ( var iIndex = 0; iIndex < oBaseArray.length; iIndex++) {
        lInfo.logln("found #" + iIndex);
        var oNewID=getNewIdValue();
        lInfo.logln("using id " + oNewID);
        try
        {
            var oTriggerNode=oBaseArray[iIndex];
            var sName = Utf8.encode(unescape(oTriggerNode.getChild('name').getValue()));
            lInfo.logln("Name: " + sName);
            var fDisabled=oTriggerNode.getChild('disabled').getValue();
            lInfo.logln("fDisabled: " + fDisabled);
            storeSubParameter(oNewID + '/name', sName);
            storeSubParameter(oNewID + '/id', oNewID);
            storeSubParameter(oNewID + '/conditions/enabled', (fDisabled == true? false : true));

            if (oTriggerNode.getChild('delay')!=null) {
                var iDelay=oTriggerNode.getChild('delay').getValue();
                lInfo.logln("iDelay: " + iDelay);
                storeSubParameter(oNewID + '/delay', iDelay);
            }

            var triggerScene=oTriggerNode.getChild('scene').getValue();
            var triggerGroup=oTriggerNode.getChild('group').getValue();
            var triggerZone=oTriggerNode.getChild('zone').getValue();
            lInfo.logln("Triggering on : " + triggerZone + " " + triggerGroup + " " + triggerScene);

            if (triggerScene == 65 || (triggerScene >= 71 && triggerScene <= 73)) {
                triggerGroup = 0;
                triggerZone = 0;
                lInfo.logln("Trigger-scene is Panik, Kommen, Gehen or Klingel -> force setting group and zone to 0");
            } else if (triggerScene >= 64) {
                triggerGroup = 0;
                lInfo.logln("Trigger-scene is apartment scene -> force setting group to 0");
            }

            var sSubpath=oNewID + '/triggers/0/';
            storeSubParameter(sSubpath + "type" ,'zone-scene');
            storeSubParameter(sSubpath + "zone" , triggerZone);
            storeSubParameter(sSubpath + "group" , triggerGroup);
            storeSubParameter(sSubpath + "scene" , triggerScene);
            storeSubParameter(sSubpath + "dsid" , -1);

            var oActionNodes=oTriggerNode.getChild('actions').getChildren();
            for (var jIndex=0;jIndex<oActionNodes.length;jIndex++) {
                var oActionNode=oActionNodes[jIndex];
                if (oActionNode.getChild('device')!=null) {
                    //device calls
                    if (oActionNode.getChild('blink')!=null) {
                        lInfo.logln("Action #" + jIndex + " on : " + oActionNode.getChild('device').getValue() + " blink");
                        storeSubParameter(oNewID + '/actions/' + jIndex + '/type', 'device-blink');
                        storeSubParameter(oNewID + '/actions/' + jIndex + '/dsid', oActionNode.getChild('device').getValue());
                    } else if (oActionNode.getChild('outputValue')!=null) {
                        lInfo.logln("Action #" + jIndex + " on : " + oActionNode.getChild('device').getValue() + " outvalue " + oActionNode.getChild('outputValue').getValue());
                        storeSubParameter(oNewID + '/actions/' + jIndex + '/type', 'device-value');
                        storeSubParameter(oNewID + '/actions/' + jIndex + '/dsid', oActionNode.getChild('device').getValue());
                        storeSubParameter(oNewID + '/actions/' + jIndex + '/value', oActionNode.getChild('outputValue').getValue());
                    } else if (oActionNode.getChild('scene')!=null){
                        lInfo.logln("Action #" + jIndex + " on : " + oActionNode.getChild('device').getValue() + " scene " +oActionNode.getChild('scene').getValue());
                        storeSubParameter(oNewID + '/actions/' + jIndex + '/type', 'device-scene');
                        storeSubParameter(oNewID + '/actions/' + jIndex + '/dsid', oActionNode.getChild('device').getValue());
                        storeSubParameter(oNewID + '/actions/' + jIndex + '/scene', oActionNode.getChild('scene').getValue());
                        storeSubParameter(oNewID + '/actions/' + jIndex + '/force', false);
                    }
                } else {
                    // zone calls
                    if (oActionNode.getChild('blink')!=null) {
                        lInfo.logln("Action #" + jIndex + " on : " + oActionNode.getChild('zone').getValue() + " " + oActionNode.getChild('group').getValue() + " blink");
                        storeSubParameter(oNewID + '/actions/' + jIndex + '/type', 'zone-blink');
                        storeSubParameter(oNewID + '/actions/' + jIndex + '/zone', oActionNode.getChild('zone').getValue());
                        storeSubParameter(oNewID + '/actions/' + jIndex + '/group', oActionNode.getChild('group').getValue());
                    } else if (oActionNode.getChild('outputValue')!=null) {
                        lInfo.logln("Action #" + jIndex + " on : " + oActionNode.getChild('zone').getValue() + " " + oActionNode.getChild('group').getValue() + " outvalue " + oActionNode.getChild('outputValue').getValue());
                        storeSubParameter(oNewID + '/actions/' + jIndex + '/type', 'zone-value');
                        storeSubParameter(oNewID + '/actions/' + jIndex + '/zone', oActionNode.getChild('zone').getValue());
                        storeSubParameter(oNewID + '/actions/' + jIndex + '/group', oActionNode.getChild('group').getValue());
                        storeSubParameter(oNewID + '/actions/' + jIndex + '/value', oActionNode.getChild('outputValue').getValue());
                    } else if (oActionNode.getChild('scene')!=null){
                        lInfo.logln("Action #" + jIndex + " on : " + oActionNode.getChild('zone').getValue() + " " + oActionNode.getChild('group').getValue() +  " scene " +oActionNode.getChild('scene').getValue());
                        storeSubParameter(oNewID + '/actions/' + jIndex + '/type', 'zone-scene');
                        storeSubParameter(oNewID + '/actions/' + jIndex + '/zone', oActionNode.getChild('zone').getValue());
                        storeSubParameter(oNewID + '/actions/' + jIndex + '/group', oActionNode.getChild('group').getValue());
                        storeSubParameter(oNewID + '/actions/' + jIndex + '/scene', oActionNode.getChild('scene').getValue());
                        storeSubParameter(oNewID + '/actions/' + jIndex + '/force', false);
                    }
                }
            }
            lInfo.logln("done. Delete old Node " + oTriggerNode.getPath());
            baseNodes.removeChild(oTriggerNode);
        }
        catch (e)
        {
            lWarn.logln("Fehler beim Datenupgrade " + e);
            if (Property.getNode('entries')!=null) {
                Property.getNode('entries').removeChild(oNewID);
            }
            fError=true;
        }
    }
    if (!fError) {
        lInfo.logln("All done. Delete old Node");
        Property.getNode('/scripts/scene-responder').removeChild('triggers');
    }
    Property.store();
}

